@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # coding: utf-8
 
-# In[138]:
+# In[134]:
 
 
 try:
@@ -12,10 +12,10 @@ except:
         pass
 
 
-# In[139]:
+# In[135]:
 
 
-version = '0.84'
+version = '0.92'
 
 import re
 import tempfile
@@ -33,7 +33,7 @@ import time
 
 # # Helper Functions
 
-# In[140]:
+# In[136]:
 
 
 def is_testing():
@@ -43,7 +43,7 @@ def is_testing():
         return False
 
 
-# In[141]:
+# In[137]:
 
 
 def isnotebook():
@@ -54,7 +54,7 @@ def isnotebook():
     return False
 
 
-# In[142]:
+# In[138]:
 
 
 def testPrint(*args):
@@ -62,6 +62,61 @@ def testPrint(*args):
         print(*args)
 
 
+# # Inspect
+
+# In[139]:
+
+
+import inspect
+def debugGetFileName():
+    return inspect.stack()[1][1]
+def debugGetLineNumber():
+    return inspect.stack()[1][2]
+def debugFunctionName():
+    return inspect.stack()[1][3]
+
+
+# In[140]:
+
+
+debugGetFileName(), debugGetLineNumber(), debugFunctionName()
+
+
+# # Shell
+
+# In[141]:
+
+
+def shell(cmd):
+    result = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+    out = result.stdout.decode()
+
+    if len(out) > 0 and out[-1] == "\n":
+        out = out[:-1]
+
+    return out
+
+
+# In[142]:
+
+
+shell_template = """/bin/bash << EOF 2>&1
+{cmd}
+EOF"""
+
+def bash(cmd, lineList=True):
+    result = subprocess.run(shell_template.format(cmd=cmd), shell=True, stdout=subprocess.PIPE)
+    out = result.stdout.decode()
+
+    if len(out) > 0 and out[-1] == "\n":
+        out = out[:-1]
+
+    if lineList:
+        return out.split('\n')
+    else:
+        return out
+
+
 # # Timing
 
 # In[143]:
@@ -90,9 +145,35 @@ if is_testing():
     print(timeSinceLastCall('test marker'))
 
 
+# In[146]:
+
+
+import time
+
+class tiTimeScope:    
+    def __init__(self, name=None):
+        if name is None:
+            name = f"{debugGetFileName()}:{debugGetLineNumber()}"
+        self.name = name
+            
+    def __enter__(self):
+        self.start = time.time()
+        return self
+
+    def __exit__(self, *args):
+        print(f"{self.name}: {time.time() - self.start:0.2f}s")
+
+if is_testing():
+    with tiTimeScope():
+        time.sleep(0.05)
+    with tiTimeScope('test'):
+        time.sleep(0.05)
+    
+
+
 # # Requiremtns
 
-# In[146]:
+# In[147]:
 
 
 requirements = []
@@ -124,7 +205,7 @@ import yaml
 
 # # Strings
 
-# In[147]:
+# In[148]:
 
 
 def sFloat0(f): return "{:0.0f}".format(f)
@@ -154,7 +235,7 @@ if is_testing():
     print(sFloatSci5(1.234567))
 
 
-# In[148]:
+# In[149]:
 
 
 def strUuid4():
@@ -165,14 +246,14 @@ def strUuid4():
     return uuid
 
 
-# In[149]:
+# In[150]:
 
 
 if is_testing():
     print(strUuid4())
 
 
-# In[150]:
+# In[151]:
 
 
 def strRand(n, num=True, upperCase=True, lowerCase=True):
@@ -184,7 +265,7 @@ def strRand(n, num=True, upperCase=True, lowerCase=True):
     return ''.join(random.choice(pool) for _ in range(n))
 
 
-# In[151]:
+# In[152]:
 
 
 if is_testing():
@@ -193,7 +274,7 @@ if is_testing():
     print(strRand(20, upperCase=False, lowerCase=False))
 
 
-# In[152]:
+# In[153]:
 
 
 def strMd5Pickle4(obj):
@@ -203,14 +284,14 @@ def strMd5Pickle4(obj):
     return md5.hexdigest()
 
 
-# In[153]:
+# In[154]:
 
 
 if is_testing():
     print(strMd5Pickle4(['a', 'b']))
 
 
-# In[154]:
+# In[155]:
 
 
 def strMd5String(s):
@@ -219,7 +300,7 @@ def strMd5String(s):
     return md5.hexdigest()
 
 
-# In[155]:
+# In[156]:
 
 
 if is_testing():
@@ -228,7 +309,7 @@ if is_testing():
 
 # # Multiprocess / Thread
 
-# In[156]:
+# In[157]:
 
 
 def mapThread(func, args, n=8):
@@ -253,14 +334,14 @@ if is_testing():
 
 # # Lists
 
-# In[157]:
+# In[158]:
 
 
 def liChunks(l, chunk_size):
     return [l[i:i + chunk_size] for i in range(0, len(l), chunk_size)]
 
 
-# In[158]:
+# In[159]:
 
 
 if is_testing():
@@ -269,7 +350,7 @@ if is_testing():
     print(liChunks(['a', 'b', 'c', 'd', 'e', 'f'], 3))
 
 
-# In[159]:
+# In[160]:
 
 
 def liSplit(a, number_of_splits):
@@ -277,7 +358,7 @@ def liSplit(a, number_of_splits):
     return [a[start:end] for start, end in zip(idxs[:-1], idxs[1:])]
 
 
-# In[160]:
+# In[161]:
 
 
 if is_testing():
@@ -289,7 +370,7 @@ if is_testing():
     print(liSplit(['a', 'b', 'c', 'd', 'e'], 6))
 
 
-# In[161]:
+# In[162]:
 
 
 def liJoin(l):
@@ -299,14 +380,14 @@ def liJoin(l):
     return out
 
 
-# In[162]:
+# In[163]:
 
 
 if is_testing():
     print(liJoin([['a', 'b', 'c'], ['d', 'e', 'f']]))
 
 
-# In[163]:
+# In[164]:
 
 
 def liFlatten(l):
@@ -323,7 +404,7 @@ def liFlatten(l):
         return l
 
 
-# In[164]:
+# In[165]:
 
 
 if is_testing():
@@ -333,7 +414,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[165]:
+# In[166]:
 
 
 regex_cache = {}
@@ -343,35 +424,35 @@ def sListFilter(l, re_include=".*", re_exclude='^\b$'):
     return [e for e in l if regex_cache[re_include].search(e) and not regex_cache[re_exclude].search(e)]
 
 
-# In[166]:
+# In[167]:
 
 
 if is_testing():
     sListFilter(['good x', 'other good x', 'without the letter before y'])
 
 
-# In[167]:
+# In[168]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[168]:
+# In[169]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[169]:
+# In[170]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[170]:
+# In[171]:
 
 
 def lCsv(l, force_oneline=False):
@@ -387,21 +468,21 @@ def lCsv(l, force_oneline=False):
     return ", ".join([str(e) for e in l])
 
 
-# In[171]:
+# In[172]:
 
 
 if is_testing():
     print(lCsv(['a', 'b', 1, 3.14]))
 
 
-# In[172]:
+# In[173]:
 
 
 if is_testing():
     print(lCsv([['a', 'b'], 1, 3.14]))
 
 
-# In[173]:
+# In[174]:
 
 
 if is_testing():
@@ -410,7 +491,7 @@ if is_testing():
 
 # # Text files
 
-# In[174]:
+# In[175]:
 
 
 def txtread(path):
@@ -431,7 +512,7 @@ def txtwrite(path, txt):
             return f.write(txt)
 
 
-# In[175]:
+# In[176]:
 
 
 if is_testing():
@@ -440,21 +521,21 @@ if is_testing():
 
 # # YAML files
 
-# In[176]:
+# In[177]:
 
 
 def yamlwrite(path, obj):
     txtwrite(path, yaml.dump(obj))
 
 
-# In[177]:
+# In[178]:
 
 
 def yamlread(path):
     return yaml.safe_load(txtread(path=path))
 
 
-# In[178]:
+# In[179]:
 
 
 if is_testing():
@@ -467,7 +548,7 @@ if is_testing():
 
 # # List of Strings
 
-# In[179]:
+# In[180]:
 
 
 regex_cache = {}
@@ -479,35 +560,35 @@ def sListFilter(listOfStrings, re_include=".*", re_exclude='^\b$'):
 
 # ### Examples
 
-# In[180]:
+# In[181]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y']))
 
 
-# In[181]:
+# In[182]:
 
 
 if is_testing():
     print(sListFilter(['good x', 'other good x', 'without the letter before y'], re_include='x'))
 
 
-# In[182]:
+# In[183]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+'))
 
 
-# In[183]:
+# In[184]:
 
 
 if is_testing():
     print(sListFilter(['not enough x', 'enough xx', 'enough xx and also z'], re_include='xx+', re_exclude='z'))
 
 
-# In[184]:
+# In[185]:
 
 
 if is_testing():
@@ -516,7 +597,7 @@ if is_testing():
 
 # # GCP Google Cloud Platform
 
-# In[185]:
+# In[186]:
 
 
 import warnings
@@ -540,7 +621,7 @@ def _gsBlob(path):
     return blob
 
 
-# In[186]:
+# In[187]:
 
 
 def gsGsToHttp(path):
@@ -550,7 +631,7 @@ def gsHttpToGs(path):
     return path.replace("https://storage.cloud.google.com/", "gs://").split("?")[0] # "?" is for possible "?authuser=0"
 
 
-# In[187]:
+# In[188]:
 
 
 def gsLs(path):
@@ -560,7 +641,7 @@ def gsLs(path):
     return ['gs://' + bucket + '/' + blob.name for blob in client.list_blobs(prefix=prefix)]
 
 
-# In[188]:
+# In[189]:
 
 
 def gsExists(path):
@@ -570,7 +651,7 @@ def gsExists(path):
     return client.blob(prefix).exists()
 
 
-# In[189]:
+# In[190]:
 
 
 def gsUploadFile(gsPath, path):
@@ -578,7 +659,7 @@ def gsUploadFile(gsPath, path):
     blob.upload_from_filename(path)
 
 
-# In[190]:
+# In[191]:
 
 
 def gsTxtread(path):
@@ -590,7 +671,7 @@ def gsTxtread(path):
     return blob.download_as_string().decode()
 
 
-# In[191]:
+# In[192]:
 
 
 def gsTxtwrite(path, s):
@@ -598,7 +679,7 @@ def gsTxtwrite(path, s):
     return blob.upload_from_string(s)
 
 
-# In[192]:
+# In[193]:
 
 
 def gsImwrite(path, img):
@@ -607,7 +688,7 @@ def gsImwrite(path, img):
     blob.upload_from_string(encimg.tobytes())
 
 
-# In[193]:
+# In[194]:
 
 
 def gsImread(path):
@@ -616,7 +697,7 @@ def gsImread(path):
     return cv2.imdecode(raw, 1)[:, :, [2, 1, 0]]
 
 
-# In[194]:
+# In[195]:
 
 
 def gsPklWrite(path, obj):
@@ -624,7 +705,7 @@ def gsPklWrite(path, obj):
     return blob.upload_from_string(pickle.dumps(obj, protocol=4))
 
 
-# In[195]:
+# In[196]:
 
 
 def gsPklRead(path):
@@ -634,7 +715,7 @@ def gsPklRead(path):
 
 # # Pickle
 
-# In[196]:
+# In[197]:
 
 
 def pklwrite(path, obj):
@@ -647,7 +728,7 @@ def pklwrite(path, obj):
             return pickle.dump(obj, f, protocol=4)
 
 
-# In[197]:
+# In[198]:
 
 
 def pklread(path):
@@ -658,7 +739,7 @@ def pklread(path):
             return pickle.load(f)
 
 
-# In[198]:
+# In[199]:
 
 
 if is_testing():
@@ -671,12 +752,20 @@ if is_testing():
 
 # # Images
 
-# In[199]:
+# In[200]:
 
 
+def _imreadSsh(path):
+    with tempfile.TemporaryDirectory() as dir_path:
+        out_path = os.path.join(dir_path, os.path.basename(path))
+        bash(f'scp {path} {out_path}')
+        return imread(out_path)
+
 def imread(path):
     if isinstance(path, np.ndarray) and path.dtype == np.uint8:
         return path
+    elif '@' in path and ':' in path:
+        return _imreadSsh(path)
     elif 'http://' in path or 'https://' in path:
         raw = np.asarray(bytearray(urlopen(path).read()), dtype="uint8")
         img = cv2.imdecode(raw, cv2.IMREAD_COLOR)
@@ -684,6 +773,8 @@ def imread(path):
         return gsImread(path)
     elif os.path.isfile(path):
         img = cv2.imread(path, cv2.IMREAD_COLOR)
+    else:
+        raise RuntimeError("Image not found: ", path)
     return img[:, :, [2, 1, 0]] # This converts the cv2 colors to RGB
 
 def imsread(paths):
@@ -700,7 +791,13 @@ def imwrite(path, img):
     
 def imjoin(imgs, axis=1, pad=0):
     if pad > 0:
-        imgs = [impad(img, right=pad) for img in imgs]
+        imgs_out = []
+        for idx, img in enumerate(imgs):
+            if idx < len(imgs):
+                imgs_out.append(impad(img, right=pad))
+            else:
+                imgs_out.append(img)
+        imgs = imgs_out
     return np.concatenate(imgs, axis=axis)
 
 def impad(img, top=0, bottom=0, left=0, right=0, color=255, mode='constant'):
@@ -712,7 +809,7 @@ def impad(img, top=0, bottom=0, left=0, right=0, color=255, mode='constant'):
         raise RuntimeError(f"Mode '{mode}' not found")
 
 
-# In[200]:
+# In[201]:
 
 
 def imscaleNN(img, s):
@@ -722,7 +819,7 @@ def imscaleBic(img, s):
     return cv2.resize(img, None, fx=s, fy=s, interpolation=cv2.INTER_CUBIC)
 
 
-# In[201]:
+# In[202]:
 
 
 def imshow(array, scale=1, force_nb=False):
@@ -737,15 +834,61 @@ def imshow(array, scale=1, force_nb=False):
         cv2.waitKey()
 
 
-# In[202]:
+# In[203]:
+
+
+def imNewNoise(height, width, scaleNN=1):
+    return imscaleNN((255 * np.random.rand(height, width, 3)).astype(np.uint8), scaleNN)
+
+
+# In[204]:
 
 
 if is_testing():
-    img = imscaleNN((255 * np.random.rand(20, 20, 3)).astype(np.uint8), 10)
+    img = imNewNoise(20, 20, 10)
     imshow(impad(img, bottom=10, right=10, mode='reflect'))
 
 
-# In[203]:
+# In[205]:
+
+
+def imCropUniformBorder(img, color=255):
+    mask = (img == color)
+    out = img.copy()
+    for axis in (0, 1):
+        const_1d = mask.mean(axis=2).mean(axis=axis)
+        diff = np.diff(const_1d)
+
+        if const_1d[0] < 1:
+            idx_start = 0
+        else:
+            idx_start = np.argmin(diff) + 1 # Diff between next and current
+
+        if const_1d[-1] < 1:
+            idx_end = 0
+        else:
+            idx_end = np.argmax(np.flip(diff)) + 1
+
+        if axis == 0:
+            out = out[:, idx_start: -idx_end or None]
+        else:
+            out = out[idx_start: -idx_end or None, :]
+    return out
+
+if is_testing():
+    # Try the imCropUniformBorder for different paddings
+    img_crop = imNewNoise(17, 5)
+    for i in range(2):
+        for j in range(2):
+            for k in range(2):
+                for l in range(2):
+                    imgPad = impad(img_crop, top=i, left=j, right=k, bottom=l)
+                    out = imCropUniformBorder(imgPad)
+                    if not (out==img_crop).all():
+                        raise RuntimeError('Crop did not work for ', i, j, k, l)
+
+
+# In[206]:
 
 
 def imDecodeBytes(imBytes):
@@ -756,7 +899,7 @@ def imDecodeBytes(imBytes):
     return img
 
 
-# In[204]:
+# In[207]:
 
 
 #if is_testing():
@@ -764,14 +907,14 @@ def imDecodeBytes(imBytes):
 #    imshow(img)
 
 
-# In[205]:
+# In[208]:
 
 
 #if is_testing():
 #    imshow('http://via.placeholder.com/70.png')
 
 
-# In[206]:
+# In[209]:
 
 
 if is_testing():
@@ -790,7 +933,7 @@ if is_testing():
         imshow(out_large_interbic)
 
 
-# In[207]:
+# In[210]:
 
 
 def imRepeat(img, height, width):
@@ -804,7 +947,7 @@ def imRepeat(img, height, width):
     return img
 
 
-# In[208]:
+# In[211]:
 
 
 if is_testing():
@@ -812,33 +955,33 @@ if is_testing():
     imshow(imRepeat(img_random_color_3x5, 7, 11), scale=10)
 
 
-# In[209]:
+# In[212]:
 
 
 def imNewWhite(height, width):
     return np.ones((height, width, 3)).astype(np.uint8) * 255
 
 
-# In[210]:
+# In[213]:
 
 
 if is_testing(): imshow(imNewWhite(10, 500) - 30)
 
 
-# In[211]:
+# In[214]:
 
 
 def imNewBlack(height, width):
     return np.zeros((height, width, 3)).astype(np.uint8)
 
 
-# In[212]:
+# In[215]:
 
 
 if is_testing(): imshow(imNewBlack(10, 500))
 
 
-# In[213]:
+# In[216]:
 
 
 def imAddNoiseGauss(img, std):
@@ -850,7 +993,7 @@ def imAddNoiseGauss(img, std):
 if is_testing(): imshow(imAddNoiseGauss(imNewWhite(150, 150), 30))
 
 
-# In[214]:
+# In[217]:
 
 
 def imJpgDegradation(img, quality):
@@ -863,7 +1006,7 @@ def imJpgDegradation(img, quality):
 #if is_testing(): imshow(imJpgDegradation(imread('http://via.placeholder.com/150.png'), 15))
 
 
-# In[215]:
+# In[218]:
 
 
 def imCropCenter(img, size):
@@ -878,35 +1021,35 @@ def imCropCenter(img, size):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[216]:
+# In[219]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 10))
 
 
-# In[217]:
+# In[220]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 49))
 
 
-# In[218]:
+# In[221]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 50))
 
 
-# In[219]:
+# In[222]:
 
 
 #if is_testing(): 
 #    imshow(imCropCenter(imread('http://via.placeholder.com/40.png'), 100))
 
 
-# In[220]:
+# In[223]:
 
 
 def imCropRandom(img, size, auto_size=True):
@@ -929,7 +1072,7 @@ def imCropRandom(img, size, auto_size=True):
     return img[h_start:h_end, w_start:w_end]
 
 
-# In[221]:
+# In[224]:
 
 
 #if is_testing(): 
@@ -939,14 +1082,27 @@ def imCropRandom(img, size, auto_size=True):
 #    imshow(imCropRandom(imread('http://via.placeholder.com/40.png'), 41))
 
 
-# In[222]:
+# In[225]:
 
 
-def imGallery(imgs, pad=0):
+def imGallery(imgs, pad=0, n_cols=None, n_rows=None):
     imgs = [imread(img) for img in imgs] # load if path or url
+    
     n = len(imgs)
-    nw = int(np.ceil(np.sqrt(n)))
-    nh = int(np.ceil(n / nw))
+    if isinstance(n_cols, int):
+        if n_rows is not None:
+            raise RuntimeError("Can only fix rows or columns")
+        nw = n_cols
+        nh = int(np.ceil(n / nw))
+    elif isinstance(n_rows, int):
+        if n_cols is not None:
+            raise RuntimeError("Can only fix rows or columns")
+        nh = n_rows
+        nw = int(np.ceil(n / nh))
+    else:
+        nw = int(np.ceil(np.sqrt(n)))
+        nh = int(np.ceil(n / nw))
+        
     img_h, img_w, _ = imgs[0].shape
     w = nw * img_w + (nw - 1) * pad
     h = nh * img_h + (nh - 1) * pad
@@ -970,7 +1126,7 @@ def imGallery(imgs, pad=0):
     return out
 
 
-# In[223]:
+# In[226]:
 
 
 if is_testing():
@@ -978,57 +1134,29 @@ if is_testing():
     listOfDummyImagesBlackToWhite = [imNewBlack(10, 10) + 25 * i for i in range(10)]
 
 
-# In[224]:
-
-
-if is_testing():
-    imshow(imGallery(listOfDummyImagesBlackToWhite[:1]))
-
-
-# In[225]:
-
-
-if is_testing():
-    imshow(imGallery(listOfDummyImagesBlackToWhite[:2]))
-
-
-# In[226]:
-
-
-if is_testing():
-    imshow(imGallery(listOfDummyImagesBlackToWhite[:2], pad=1))
-
-
 # In[227]:
 
 
 if is_testing():
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:1]))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:2]))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:2], pad=1))
     imshow(imGallery(listOfDummyImagesBlackToWhite[:5], pad=1))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:9], pad=1))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1, n_cols=3))
+    imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1, n_rows=2))
 
 
 # In[228]:
 
 
-if is_testing():
-    imshow(imGallery(listOfDummyImagesBlackToWhite[:9], pad=1))
-
-
-# In[229]:
-
-
-if is_testing():
-    imshow(imGallery(listOfDummyImagesBlackToWhite[:10], pad=1))
-
-
-# In[230]:
-
-
 #if is_testing():
 #    img40x20 = imread("http://via.placeholder.com/40x20.png")
 #    imshow(imGallery([img40x20, img40x20, img40x20], pad=1))
 
 
-# In[231]:
+# In[229]:
 
 
 def imAddTextHeader(img, text, h=None, size=None):
@@ -1041,7 +1169,7 @@ def imAddTextHeader(img, text, h=None, size=None):
     
     text_img = np.ones((h, img.shape[1], img.shape[2]), dtype=img.dtype) * 255
     im = Image.fromarray(text_img)
-
+    
     draw = ImageDraw.Draw(im)
     
     w = text_img.shape[1]
@@ -1049,8 +1177,8 @@ def imAddTextHeader(img, text, h=None, size=None):
     centered=True
     middle = True
     text_width, text_height = draw.textsize(text)
-    draw.text((w/2 - (text_width/2) * centered, text_height/2 - h/2)
-        ,text, fill="black")
+    # Top left corner of text (w, h)
+    draw.text((w/2 - (text_width/2) * centered, h/2 - text_height/2), text, fill="black")
     
     text = np.array(im)[:, :, :3]
     
@@ -1060,12 +1188,12 @@ def imAddTextHeader(img, text, h=None, size=None):
 
 if is_testing():
     imshow(imAddTextHeader(imNewBlack(100, 100), 'black image'))
-    imshow(imAddTextHeader(imNewBlack(100, 100), 'black image', h = 20))
+    imshow(imAddTextHeader(imNewBlack(100, 100), 'black image'))
 
 
 # # Video
 
-# In[232]:
+# In[230]:
 
 
 #def imgsToVid(imgs, file_name='out.avi'):
@@ -1087,7 +1215,7 @@ if is_testing():
 
 # # pytorch
 
-# In[233]:
+# In[231]:
 
 
 def ptT(img_np_uint8):
@@ -1104,7 +1232,7 @@ def ptUint8(img_tensor):
     
 
 
-# In[234]:
+# In[232]:
 
 
 if is_testing():
@@ -1116,14 +1244,21 @@ if is_testing():
 
 # # Files
 
-# In[235]:
+# In[233]:
 
 
+def _fiFindByWildcardSsh(wildcard):
+    login, path = wildcard.split(':')
+    paths = bash(f"ssh {login} 'ls {path}'")
+    return [f'{login}:{p}' for p in natsort.natsorted(paths)]
+
 def fiFindByWildcard(wildcard):
+    if '@' in wildcard and ':' in wildcard:
+        return _fiFindByWildcardSsh(wildcard)
     return natsort.natsorted(glob.glob(wildcard, recursive=True))
 
 
-# In[236]:
+# In[234]:
 
 
 if is_testing():
@@ -1152,7 +1287,7 @@ if is_testing():
         imshow(imGallery(imgs))
 
 
-# In[237]:
+# In[235]:
 
 
 def fiRemove(path):
@@ -1180,7 +1315,7 @@ def fiAgeHours(path):
     return (time.time() - os.stat(path)[stat.ST_MTIME]) / 3600
 
 
-# In[238]:
+# In[236]:
 
 
 if is_testing():
@@ -1212,7 +1347,7 @@ if is_testing():
         print("files: " + "\n".join(fiFindByWildcard(os.path.join(d, "**", "*"))))
 
 
-# In[239]:
+# In[237]:
 
 
 def fiUnzip(zip_path, out_dir=None, remove=False):
@@ -1227,7 +1362,7 @@ def fiUnzip(zip_path, out_dir=None, remove=False):
         zip.extractall(out_dir)
 
 
-# In[240]:
+# In[238]:
 
 
 def fiJoin(*args):
@@ -1239,7 +1374,7 @@ if is_testing():
 
 # # Jupyter
 
-# In[241]:
+# In[239]:
 
 
 def jnSettings():
@@ -1250,7 +1385,7 @@ if is_testing():
     jnSettings()
 
 
-# In[242]:
+# In[240]:
 
 
 def jnNbFullWidth():
@@ -1258,13 +1393,13 @@ def jnNbFullWidth():
     display(HTML("<style>.container { width:100% !important; }</style>"))
 
 def jnPandasSettings():
-    pd.set_option('display.max_rows', 1000)
-    pd.set_option('display.max_columns', 100)
-    pd.set_option('display.width', 1000)
+    pd.set_option('display.max_rows', 100000)
+    pd.set_option('display.max_columns', 10000)
+    pd.set_option('display.width', 100000)
     pd.set_option('display.max_colwidth', 100000)
 
 
-# In[243]:
+# In[241]:
 
 
 def jnImageRenderingPixelated():
@@ -1276,7 +1411,7 @@ def jnImageRenderingAuto():
     display(HTML('<style type = text/css> img { image-rendering: auto; } </style>'))
 
 
-# In[244]:
+# In[242]:
 
 
 def jnDefault():
@@ -1294,7 +1429,7 @@ if is_testing():
 
 
 
-# In[245]:
+# In[243]:
 
 
 def jnHtmlRow(strList, width=100):
@@ -1307,7 +1442,7 @@ if is_testing():
     
 
 
-# In[246]:
+# In[244]:
 
 
 def jnImgToHtmlB64(img):
@@ -1342,7 +1477,7 @@ def jnImageRow(imgs, names=None, width=100, external=False, title_img=None, titl
     display(HTML(html))
 
 
-# In[247]:
+# In[245]:
 
 
 if is_testing():
@@ -1350,7 +1485,7 @@ if is_testing():
    
 
 
-# In[248]:
+# In[246]:
 
 
 if is_testing():
@@ -1360,7 +1495,7 @@ if is_testing():
     jnImageRow(imgs, external=True, title='test')
 
 
-# In[249]:
+# In[247]:
 
 
 
@@ -1368,44 +1503,9 @@ if is_testing():
    jnImageRow(imgs=[imNewBlack(20, 20), imNewBlack(40, 20), imNewBlack(20, 800)])
 
 
-# # Shell
-
-# In[250]:
-
-
-def shell(cmd):
-    result = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
-    out = result.stdout.decode()
-
-    if len(out) > 0 and out[-1] == "\n":
-        out = out[:-1]
-
-    return out
-
-
-# In[251]:
-
-
-shell_template = """/bin/bash << EOF 2>&1
-{cmd}
-EOF"""
-
-def bash(cmd, lineList=True):
-    result = subprocess.run(shell_template.format(cmd=cmd), shell=True, stdout=subprocess.PIPE)
-    out = result.stdout.decode()
-
-    if len(out) > 0 and out[-1] == "\n":
-        out = out[:-1]
-
-    if lineList:
-        return out.split('\n')
-    else:
-        return out
-
-
 # # GIT
 
-# In[252]:
+# In[248]:
 
 
 def gitGetHash():
@@ -1415,7 +1515,7 @@ if is_testing():
     print(gitGetHash())
 
 
-# In[253]:
+# In[249]:
 
 
 def gitGetBranch():
@@ -1425,7 +1525,7 @@ if is_testing():
     print(gitGetBranch())
 
 
-# In[254]:
+# In[250]:
 
 
 def gitGetMessage():
@@ -1435,7 +1535,7 @@ if is_testing():
     print(gitGetMessage())
 
 
-# In[255]:
+# In[251]:
 
 
 def gitGetAllFiles():
@@ -1445,7 +1545,7 @@ if is_testing():
     print(gitGetAllFiles())
 
 
-# In[256]:
+# In[252]:
 
 
 def gitListByTime(path=None):
@@ -1461,7 +1561,7 @@ if is_testing():
     print(gitListByTime()[-3:])
 
 
-# In[257]:
+# In[253]:
 
 
 def gitTime(path):
@@ -1472,7 +1572,7 @@ if is_testing():
     print(gitTime("__init__.py"))
 
 
-# In[258]:
+# In[254]:
 
 
 def gitTimeRaw(path):
@@ -1483,7 +1583,7 @@ if is_testing():
     print(gitTimeRaw("__init__.py"))
 
 
-# In[259]:
+# In[255]:
 
 
 def shellSortDirsBySize():
@@ -1495,14 +1595,14 @@ if is_testing():
 
 # # Package Files
 
-# In[260]:
+# In[256]:
 
 
 if is_testing():
     package_files = {}
 
 
-# In[261]:
+# In[257]:
 
 
 def shell(cmd):
@@ -1510,7 +1610,7 @@ def shell(cmd):
     return result.stdout.decode()
 
 
-# In[262]:
+# In[258]:
 
 
 if is_testing():
@@ -1542,7 +1642,7 @@ setup(
 )""".format(requirements=", ".join(["'" + r + "'" for r in requirements]), version=version)
 
 
-# In[263]:
+# In[259]:
 
 
 if is_testing():
@@ -1552,7 +1652,7 @@ description-file = README.md
 """
 
 
-# In[264]:
+# In[260]:
 
 
 if is_testing():
@@ -1575,23 +1675,22 @@ OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE."""
 
 
-# In[265]:
+# In[268]:
 
 
 if is_testing():
     package_files['README.md'] = ""
 
 
-# In[266]:
+# In[269]:
 
 
 if is_testing():
     shell("jupyter nbconvert --to script lx.ipynb")
     package_files['lx/__init__.py'] = txtread("lx.py")
-    
 
 
-# In[267]:
+# In[270]:
 
 
 if is_testing():
@@ -1602,7 +1701,7 @@ lx/__init__.py
 """
 
 
-# In[268]:
+# In[271]:
 
 
 if is_testing():
@@ -1617,7 +1716,7 @@ if is_testing():
         
         if output.strip() == version:
             #print(shell("cd {d} && git status && git config user.name 'lx' && git config user.email 'lx' && git commit -a -m 'Add' && git log && git push && python setup.py sdist && twine upload dist/* --verbose".format(d=d)))
-            print(shell("cd {d} && python3 setup.py sdist && twine upload dist/* --verbose".format(d=d)))
+            print("\n".join(bash("cd {d} && python setup.py sdist && ~/miniconda3/bin/python -m twine upload dist/* --verbose".format(d=d))))
 
 
 # In[ ]:
